""" Main application for the API wrapper """
import logging
import os

import requests

from metofficeamd.constants import DOMAIN, ROOT
from metofficeamd.models import FileDetails, OrderDetails, OrderList, RunList, RunListForModel

logger = logging.getLogger(__name__)


class BaseMetOfficeAMD:
    """Main class for connection and retrieving data from Met Office Weather DataHub AMD"""

    def __init__(
        self,
        cache_dir: str = "./temp_metofficeamd",
        client_id: str = None,
        client_secret: str = None,
    ):
        """
        Initialise the class

        :param cache_dir: The directory were files are downloaded to
        :param client_id: the client id for the api
        :param client_secret: the client secret for the api
        """

        if client_id is None:
            self.client_id = os.environ["API_KEY"]
        else:
            self.client_id = client_id

        if client_secret is None:
            self.client_secret = os.environ["API_SECRET"]
        else:
            self.client_secret = client_secret

        self.make_headers()

        self.cache_dir = cache_dir

    def make_headers(self):
        """
        Make header object
        """
        logger.debug("setting headers to call api")

        self.headers = {
            "X-IBM-Client-Id": self.client_id,
            "X-IBM-Client-Secret": self.client_secret,
            "accept": "application/json",
        }

    def call_url(self, url: str, headers: dict = None) -> requests.Response:
        """
        Call url string using request library.

        :param url: url to be called
        :return: response from url
        """
        if headers is None:
            headers = self.headers

        url = f"{url}?detail=MINIMAL"
        logger.debug(f"Calling url {url}")

        response = requests.get(url, headers=headers)

        # check response code 200 and show error if not
        logger.debug(response.status_code)
        assert response.status_code == 200

        return response

    def get_orders(self) -> OrderList:
        """Get a list of order"""

        response = self.call_url(url=f"https://{DOMAIN}/{ROOT}/orders")

        data = response.json()

        return OrderList(**data)

    def get_lastest_order(self, order_id) -> OrderDetails:
        """
        Provide a list of the latest available data files for the specified order.

        :param order_id: The order ID that you wish to retrieve information about. The Order ID can
            be seen under a specific order on the Atmospheric Weather Data Tool Order Summary Page
            or found in the list of orders in the JSON response from your call to /1.0.0/orders
        :return: The latest order
        """

        response = self.call_url(url=f"https://{DOMAIN}/{ROOT}/orders/{order_id}/latest")

        data = response.json()["orderDetails"]

        return OrderDetails(**data)

    def get_latest_order_file_id(self, order_id, file_id) -> FileDetails:
        """
        Provide the details of a specific file that can be obtained for the latest available data.

        :param order_id: The order ID that you wish to retrieve information about. The Order ID can
            be seen under a specific order on the Atmospheric Weather Data Tool Order Summary Page
            or found in the list of orders in the JSON response from your call to /1.0.0/orders
        :param file_id: The file ID of the application/x-grib file you wish to retrieve information
            about. The file IDs can be seen on the Atmospheric Weather Data Tool Order Summary Page
             or found in the JSON response from your call to /1.0.0/orders/{orderId}/latest
        :return: Pydantic object of the details of the file
        """

        response = self.call_url(url=f"https://{DOMAIN}/{ROOT}/orders/{order_id}/latest/{file_id}")

        data = response.json()["fileDetails"]
        print(data)

        return FileDetails(**data)

    def get_lastest_order_file_id_data(self, order_id, file_id, filename: str = None) -> str:
        """
        Gets the actual data for a specific file that can be obtained for the latest available data.

        :param order_id: The order ID that you wish to retrieve information about. The Order ID can
            be seen under a specific order on the Atmospheric Weather Data Tool Order Summary Page
            or found in the list of orders in the JSON response from your call to /1.0.0/orders
        :param file_id: The file ID of the application/x-grib file you wish to retrieve information
            about. The file IDs can be seen on the Atmospheric Weather Data Tool Order Summary Page
             or found in the JSON response from your call to /1.0.0/orders/{orderId}/latest
        :param filename: the name of the file that will be saved
        :return: filename where the data is downloaded to
        """

        headers = self.headers.copy()
        headers["accept"] = "application/x-grib"

        if filename is None:
            filename = f"{order_id}_{file_id}.grib"

        filename = f"{self.cache_dir}/{filename}"
        if not os.path.exists(filename):

            data = self.call_url(
                url=f"https://{DOMAIN}/{ROOT}/orders/{order_id}/latest/{file_id}/data",
                headers=headers,
            )

            if not os.path.isdir(self.cache_dir):
                os.mkdir(self.cache_dir)

            with open(filename, mode="wb") as localfile:
                localfile.write(data.content)

        return filename

    def get_runs(self) -> RunList:
        """
        List all runs

        :return: pydantic object of run list
        """

        response = self.call_url(url=f"https://{DOMAIN}/{ROOT}/runs")

        data = response.json()

        return RunList(**data)

    def get_runs_model_id(self, model_id) -> RunListForModel:
        """
        List all runs for specific model

        :param model_id: the model id we are looking for
        :return: Pydantic object of specific run list for a model
        """

        response = self.call_url(url=f"https://{DOMAIN}/{ROOT}/runs/{model_id}")

        data = response.json()

        return RunListForModel(**data)



""" Pydantic models for the return objects of the API """
from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel


class Extent(BaseModel):
    """Extent model"""

    t: List[datetime]
    z: List[int]


class OrderInfo(BaseModel):
    """Order info model"""

    orderId: str
    name: str
    modelId: str
    requiredLatestRuns: Optional[List[int]]
    format: str


class File(BaseModel):
    """File model"""

    fileId: str
    runDateTime: datetime
    run: int
    local_filename: Optional[str]
    timesteps: Optional[List[int]]


class OrderList(BaseModel):
    """Order list model"""

    orders: List[OrderInfo] = []


class OrderDetails(BaseModel):
    """Order details model"""

    order: OrderInfo
    files: List[File]


class RunDetails(BaseModel):
    """Run details model"""

    run: int
    runDateTime: datetime
    runFilter: str


class RunListForModel(BaseModel):
    """Run list for a specific model"""

    modelId: str
    completeRuns: List[RunDetails]


class RunList(BaseModel):
    """Run list model"""

    runs: List[RunListForModel]


class ParameterDetails(BaseModel):
    """Parameter details model"""

    parameterId: str
    extent: Extent


class FileDetails(BaseModel):
    """File details model"""

    parameterDetails: List[ParameterDetails]
    file: File



import logging
from typing import List, Optional

import cfgrib
import xarray as xr

from metofficeamd.base import BaseMetOfficeAMD

logger = logging.getLogger(__name__)

VARS_TO_DELETE = (
    "unknown",
    "valid_time",
    "heightAboveGround",
    "heightAboveGroundLayer",
    "atmosphere",
    "cloudBase",
    "surface",
    "meanSea",
    "level",
)


class MetOfficeAMD(BaseMetOfficeAMD):
    """Class built on top of BaseMetOfficeAMD used for processing multiple files"""

    def download_all_files(self, order_ids: Optional[List[str]] = None):
        """Download all files in the latest"""

        if order_ids is None:
            all_orders = self.get_orders()
            order_ids = [order.orderId for order in all_orders.orders]

        # loop over orders
        self.files = []
        for order_id in order_ids:

            logger.debug(f"Loading files from order {order_id}")

            self.order_details = self.get_lastest_order(order_id=order_id)

            # loop over all files
            for file in self.order_details.files:
                file_id = file.fileId

                variable = file.fileId
                datetime = variable.split("_")[-2]

                # There seems to be two files that are the same, one with '+HH' and one with 'YYYYMMDDHH'
                if datetime[0] != "+":

                    # download file
                    filename = self.get_lastest_order_file_id_data(
                        order_id=order_id, file_id=file_id
                    )

                    # put local file in file object
                    file.local_filename = filename
                    self.files.append(file)

    def load_file(self, file) -> xr.Dataset:
        """Load one grib file"""

        datasets_from_grib: list[xr.Dataset] = cfgrib.open_datasets(file)

        merged_ds = xr.merge(datasets_from_grib)

        return merged_ds

    def load_all_files(self) -> xr.Dataset:
        """Load all files and join them together"""

        # loop over all files and load them
        all_datasets_per_filename = {}
        for file in self.files:
            variable = file.fileId
            variable = variable.split("_")[1]

            dataset = self.load_file(file=file.local_filename)
            if variable not in all_datasets_per_filename.keys():
                all_datasets_per_filename[variable] = [dataset]
            else:
                all_datasets_per_filename[variable].append(dataset)

        # loop over different variables and join them together
        all_dataset = []
        for k, v in all_datasets_per_filename.items():

            # join all variables toegther
            dataset = xr.concat(v, dim="step")

            # remove un-needed variables
            for var in VARS_TO_DELETE:
                if var in dataset.variables:
                    del dataset[var]

            all_dataset.append(dataset)

        dataset = xr.merge(all_dataset)

        return dataset
